import collections
from itertools import chain
import pandas as pd
import numpy as np
from sklearn.model_selection import StratifiedKFold, KFold, \
                                    StratifiedShuffleSplit, learning_curve
from sklearn.metrics import confusion_matrix, make_scorer


def train_and_score(model_obj, score_funcs, X_train, y_train,
                     X_test, y_test, train_scores=True):
    """
    Trains model to training data, outputs score(test data) 
    for each score in "score_funcs", and does the same for 
    train data unless "train_data" is set to False
    """
    def apply_score_func(func):
        if train_scores is True:
            return (func(y_train, y_hat_train), func(y_test, y_hat_test))
        else:
            return func(y_test, y_hat_test)

    model = model_obj.fit(X_train, y_train)

    if train_scores is True:
        y_hat_train = model.predict(X_train)
    y_hat_test = model.predict(X_test)

    return list(map(apply_score_func, score_funcs))


def cv_engine(X, y, model_obj, score_funcs, splits=5, scale_obj=None, 
              train_scores=True, random_state=0, cross_validator_obj=None):
    """
    Splits data (based on whether model is classifier
    or regressor) and passes each fold to the `train_and_score`
    function.

    Collects results and returns results as `list`.

    parameters
    ----------

    X: pd.DataFrame
        Exogenous variables to be used as model inputs.

    y: pd.Series
        Endogenous variable that should be predicted by the model.

    model_obj: sklearn.BaseEstimator
        Instantiated (i.e. this is an instance of a class to which 
        hyper-parameters have already been passed) scikit-learn model, or model 
        with similar API - i.e.`model_obj.fit(X, y)` and `model_obj.predict(X)` 
        would be used for model fitting and predicting, respectively. E.g. 
        `sklearn.RandomForestClassifier(max_depth=16)`.

    score_funcs: list-like of callables
        Score functions to be run on model predictions. E.g.
        `[sklearn.metrics.f1_score, sklearn.metrics.accuracy_score]`.

    splits: int, optional(default=5)
        Number of splits to use for k-fold cross-validation.

    scale_obj: sklearn.TransformerMixin (default=None)
        Should be a scikit-learn transform object (i.e. inherits from 
        `sklearn.BaseEstimator` and `sklearn.TransformerMixin`) or have 
        a similar API (i.e. `scale_obj.fit_transform()` works as expected). 
        If passed, X will be scaled within each fold so as to prevent data 
        leakage. 

    train_scores: bool (default=True)
        Determines whether training scores, in addition to test scores, are 
        returned. Train scores are useful for comparing to test scores in order 
        to assess model fit.

    `random_state`: int, RandomState instance or None (default=0)
        Set the seed used for random sampling in cross-validation, which allows 
        for reproducability of data splits. See sklearn docs, e.g. for 
        `StratifiedKFold`, for more information.

    `cross_validator_obj`: `sklean.model_selection._BaseKFold`
        Can pass a cross-validator to override defaul CV behavior (e.g. when
        geting results for a learning curve).

    return
    ------

    `list` of `train_and_score` results, with length `splits`.

    """
    if cross_validator_obj is not None:
        skf = cross_validator_obj

    elif model_obj._estimator_type == "classifier":
        skf = StratifiedKFold(n_splits=splits, shuffle=True,
                              random_state=random_state)

    elif model_obj._estimator_type == "regressor":
        skf = KFold(n_splits=splits, shuffle=True, random_state=random_state)

    else:
        raise TypeError("Improper model type.")

    results = []
    for train, test in skf.split(X, y):
        y_train = y.iloc[train]
        y_test = y.iloc[test]

        X_train = X.iloc[train, :]
        X_test = X.iloc[test, :]

        if scale_obj is not None:
            X_train = scale_obj.fit_transform(X_train)
            X_test = scale_obj.fit_transform(X_test)

        results.append(
            train_and_score(
                    model_obj, score_funcs, 
                    X_train, y_train,
                    X_test, y_test,
                    train_scores
                ))

    return results


def format_cv_results(results, score_funcs, train_scores=True):
    """
    Takes results from cv_engine and returns as 
    unaggregated DataFrame, where trial number & score 
    function used are represented in index.
    """
    if train_scores is False:
        cols = [score.__name__ for score in score_funcs]
        return pd.DataFrame(results, columns=cols)

    else:
        res = np.array([tuple(chain(*trial)) for trial in results])
        dfs = []

        for i in range(0, res.shape[1], 2):
            dfs.append(
                pd.concat(
                    {
                    score_funcs[i // 2].__name__: 
                        pd.DataFrame(res[:, i:i+2], columns=["train", "test"])
                        },
                    axis=1
                    ))

        return dfs[0].join(dfs[1:])


def describe_dataframe(results, stats_to_run=["mean", "std"]):
    """
    Given DF of CV results, returns DF of descriptive statistics.

    parameters
    ----------

    results: pandas.DataFrame
        CV results, upon upon which descriptive statistics are to be run.

    stats_to_run: str or list-like of str
        pandas.DataFrame method name(s) indicating statistic to be run,
        e.g. "mean" or ["mean", "std"]

    return
    ------

    pandas.DataFrame of descriptive statistics, where rows correspond to
    columns of `results` and where each columns correspond to `stats_to_run`.

    """
    if isinstance(stats_to_run, str):
        stats_to_run = [stats_to_run]

    get_stats = lambda func_name: getattr(results, func_name)(
                                         ).rename(func_name
                                         ).to_frame()

    to_return = [get_stats(s) for s in stats_to_run]

    return to_return[0].join(to_return[1:])


def cv_score(X, y, model_obj, score_funcs, stats_to_run=["mean", "std"],
            train_scores=True, **cv_engine_kwargs):
    """
    Cross-validates passed model and returns performance statistics. 
    Cross-validiation is performed using shuffling. If passed model is a 
    classifier, shuffling is performed such that existing stratification 
    of classes is preserved.

    parameters
    ----------

    X: pd.DataFrame
        Exogenous variables to be used as model inputs.

    y: pd.Series
        Endogenous variable that should be predicted by the model.

    model_obj: sklearn.BaseEstimator
        Instantiated (i.e. this is an instance of a class to which 
        hyper-parameters have already been passed) scikit-learn model, 
        or model with similar API - i.e. `model_obj.fit(X, y)` and 
        `model_obj.predict(X)` would be used for model fitting and predicting, 
        respectively. E.g. `sklearn.RandomForestClassifier(max_depth=16)`.

    score_funcs: callable, or list-like of callables
        Score function(s) to be run on model predictions. E.g. 
        `sklearn.metrics.accuracy_score` or 
        `[sklearn.metrics.f1_score, sklearn.metrics.accuracy_score]`.

    stats_to_run: str or list-like of str, optional(default=["mean", "std"])
        pandas.DataFrame method name(s) indicating statistic to be run,
        e.g. "mean" or `["mad", "var"]`.

    train_scores: bool (default=True)
        Determines whether training scores, in addition to test scores, are 
        returned. Train scores are useful for comparing to test scores in order 
        to assess model fit.

    See `cv_engine` for more information on additional kwargs.

    return
    ------

    pandas.DataFrame of descriptive statistics, as specified in 
    `describe_dataframe`.
    """
    if callable(score_funcs):
        score_funcs = [score_funcs]

    return describe_dataframe(
                format_cv_results(
                    cv_engine(
                        X, y, model_obj, score_funcs, train_scores=train_scores,
                        **cv_engine_kwargs
                        ),
                    score_funcs, train_scores))


def cv_conf_mat(X, y, model_obj, **cv_engine_kwargs):
    """
    Return confusion matrix for each CV trial.

    See `cv_engine` for more information on additional kwargs.
    """
    results = cv_engine(X=X, y=y, model_obj=model_obj, 
                        score_funcs=[confusion_matrix],
                        train_scores=False,
                        **cv_engine_kwargs)

    results = [pd.concat(
                {
                    i: pd.DataFrame(trial[0],
                                    index=["neg_true", "pos_true"],
                                    columns=["neg_pred", "pos_pred"])
                }
               ) for i, trial in enumerate(results, 1)]

    return pd.concat(results)


def validate_param_range(X, y, model_type, param_name, param_range,
                         score_funcs, other_params={}, train_scores=True,
                         **cv_engine_kwargs):
    """
    Returns `validation.cv_score` across values in `param_range`
    for `param_name`, which should be a working parameter for the
    passed model.

    `model_type` should be an uninstantiated sklearn model (or
    one with similar fit and predict methods). Additional 
    hyper-parameters (i.e. not `param_name` should be passed
    in to `other_params` as dictionary.

    Please see `validation.cv_engine` for details on other args.
    """ 
    results = {}
    for val in param_range:
        model_obj = model_type(**{param_name: val}, **other_params)

        some_kwargs = {"model_obj": model_obj, "X": X, "y": y, 
                       **cv_engine_kwargs}

        other_kwargs = {"train_scores": train_scores,
                        "score_funcs": score_funcs}

        if isinstance(val, collections.Iterable):
            val = str(val)

        res = cv_engine(**some_kwargs, **other_kwargs)
        results[val] = format_cv_results(res, **other_kwargs)

    to_return = pd.concat(results)
    to_return.index = to_return.index.rename(param_name, level=0)

    return to_return


def validate_train_sizes(X, y, model_obj, score_funcs, splits=5,
                         train_sizes=[0.1, 0.33, 0.55, 0.78, 1.],
                         scale_obj=None, random_state=0):
    """
    Cross validates 'model_obj' over different `train_sizes`, which allows 
    for both insight into model fit and insight into whether model might 
    benefit from access to more data.

    Please see `validation.cv_engine` for details on other args.
    """
    if callable(score_funcs):
        score_funcs = [score_funcs]

    results = {}
    for size in train_sizes:
        train_size = (1 - (1 / splits)) * size

        cross_validator_obj = StratifiedShuffleSplit(n_splits=splits,
                                                     train_size=train_size,
                                                     test_size=1 / splits,
                                                     random_state=random_state)

        some_kwargs = {"model_obj": model_obj, "X": X, "y": y, 
                       "splits": splits, "scale_obj": scale_obj,
                       "cross_validator_obj": cross_validator_obj,
                       "random_state": random_state}

        other_kwargs = {"train_scores": True, "score_funcs": score_funcs}

        res = cv_engine(**some_kwargs, **other_kwargs)
        results[size] = format_cv_results(res, **other_kwargs)

    to_return = pd.concat(results)
    return to_return


import numpy as np
import pandas as pd
from functools import partial, update_wrapper
from sklearn.metrics import confusion_matrix

from indoorplants.validation import crossvalidate


def _score_prob_as_class(t, func):
    """Returns modified 'func' that is ready to take
    a vector of class probabilities and return a 
    the results of a score function that relies on
    class labels."""
    def threshold(t, y_true, y_pred):
        pos_class = y_pred[:, 1]
        conv = np.vectorize(lambda _: 1 if _ > t else 0)
        return func(y_true, conv(pos_class))

    def wrap(t):    
        partial_func = partial(threshold, t)
        update_wrapper(partial_func, func)
        return partial_func
    
    return wrap(t)


def _convert_score_funcs(t, score_funcs):
    """Takes an iterable of scoring functions and returns
    modified version of each function, modified by 
    _score_prob_as_class."""
    new_funcs = []
    for score in score_funcs:
        new_funcs.append(_score_prob_as_class(t, score))
    return new_funcs


def cv_score(X, y, model_obj, score_funcs, thresholds, 
             splits=5, train_scores=True, scale_obj=None):
    """Performs cross validation of passed 'model' for the
    different decision boundaries passed in 'thresholds',
    which must be an iterable.

    Returns crossvalidate.cv_score. Please see this function's 
    docs for more information on other arguments."""
    model_obj.predict = model_obj.predict_proba
    results = {}
    for t in thresholds:
        scores = _convert_score_funcs(t, score_funcs)
        kwargs = {'X':X, 'y':y, 'model_obj':model_obj, 
                  'splits':splits, 'scale_obj':scale_obj, 
                  'score_funcs':scores, 
                  'train_scores': train_scores}
        results[t] = crossvalidate.cv_score(**kwargs)
    return pd.concat(results)


def cv_conf_mat(X, y, model_obj, t, splits=3, scale_obj=None):
    """Returns confusion matrix resulting from 'splits'-fold
    cross-validation using 't' to determine decision boundary."""
    model_obj.predict = model_obj.predict_proba
    scores = _convert_score_funcs(t, [confusion_matrix])
    
    results = crossvalidate.cv_engine(X=X, y=y, model_obj=model_obj, 
                                     score_funcs=scores, 
                                     splits=splits, scale_obj=scale_obj, 
                                     train_scores=False) 
    results = [pd.concat({i: pd.DataFrame(trial[0],
                            index=['neg_true', 'pos_true'],
                            columns=['neg_pred', 'pos_pred'])}) 
                   for i, trial in enumerate(results, 1)]

    return pd.concat({t: pd.concat(results)})


import numpy as np
import scipy.stats as sp
import matplotlib.pyplot as plt

from . import features

def qq_plot(series, figsize=(11, 8)):
    """compare given series to Normal distirbution
    with matching location & scale"""
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    arrs, res = sp.probplot(series, 
                            dist=sp.norm, 
                            sparams=(series.mean(), series.std()), 
                            plot=ax)
    bbox = {'fc': '1', 'pad': 3}
    xy = (ax.get_xticks()[-2], ax.get_yticks()[2])
    text = ax.annotate(r'$R^2$: {}'.format(round(res[2], 3)),
                     xy=xy,
                     bbox=bbox)
    title = ax.set_title('Q-Q plot')
    return ax


def feature_value_counts_by_class(table, class_col, feature, 
                                  figsize=(11, 8), **kwargs):
    """plot histogram of feature, with data color-coded by class"""
    ax = table.groupby([class_col, feature]
             ).size(
             ).unstack(0
             ).plot.bar(stacked=True, figsize=figsize, 
                        alpha=.5, **kwargs)
    
    plt.legend(loc='best')
    title = plt.title('{} histogram, across {}'.format(class_col, feature))
    return ax


def feature_hist_by_class(eda_df, class_col, feature, bins=None): 
    """
    Plots a histogram of passed feature, broken out by class, which
    must be of a binary nature.
    """
    fig = plt.figure(figsize=(11, 8))
    ax = fig.add_subplot(111)
    ax = eda_df[eda_df[class_col] == 1
               ][feature].hist(ax=ax, bins=bins, color="orange", alpha=.5)
    ax = eda_df[eda_df[class_col] == 0
               ][feature].hist(ax=ax, bins=bins, color="blue", alpha=.5)
    l = ax.set_ylabel("count")
    l = ax.set_xlabel(f"{feature}")
    t = ax.set_title(f"hist: {feature}, by {class_col} (orange=1, blue=0)")
    return ax


def classes_across_feature(table, class_col, feature, figsize=(11, 8)):
    """scatter feature against class"""
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    classes = table[class_col].unique()
    filtered = table.loc[table[class_col] == classes[0], :]
    ax.scatter(x=filtered[feature], y=filtered[class_col], label=classes[0])

    for i, val in enumerate(classes[1:], 1):
        filtered = table.loc[table[class_col] == val, :]
        ax.scatter(x=filtered[feature],y=filtered[class_col],
                  label=val, color='C{}'.format(i))

    plt.yticks([classes[0], classes[-1]])
    plt.legend(loc='best')
    title = plt.title('{}: by {}'.format(class_col, feature))
    return ax


def scatter_by_class(table, class_col, x, y, figsize=(11, 8), alpha=.5):
    """scatter two features, color code by class"""
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    classes = table[class_col].unique()
    filtered = table.loc[table[class_col] == classes[0], :]
    ax.scatter(x=filtered[x], y=filtered[y], alpha=alpha, label=classes[0])

    for i, val in enumerate(classes[1:], 1):
        filtered = table.loc[table[class_col] == val, :]
        ax.scatter(x=filtered[x], y=filtered[y], alpha=alpha, label=val,
                   color='C{}'.format(i))

    ax.set_xlabel(x)
    ax.set_ylabel(y)
    plt.legend(loc='best')
    title = plt.title('{}: {} vs. {}'.format(class_col, y, x))
    return ax


def box_plot_by_class(eda_df, class_col, feature):
    
    ax = eda_df[[f"{feature}", f"{class_col}"]
              ].pivot(values=f"{feature}", columns=f"{class_col}"
              ).boxplot(figsize=(11, 8))
    l= ax.set_xlabel(f"{class_col}")
    l= ax.set_ylabel(f"{feature}")
    t = ax.set_title(f"box plot: {feature} by {class_col}")
    return ax


def center_scale_plot(series, center_func, scale_func, bins=None,
                      figsize=(11, 8), return_bins=False, ndigits=4):
    """produce histogram overlayed with bands corresponding to
    units of scale from the center. user supplies center and 
    scale functions"""
    
    # create figure, plot hist, and get max dimensions
    if series.dtype not in (int, float):
        raise TypeError("Must pass numeric data for histogram!")

    plt.figure()
    if bins is None:
        bins = max(min(int(round(series.nunique(), -2) / 10), 100), 10)

    ax = series.hist(figsize=figsize, bins=bins)

    ymaxtick = ax.get_yticks()[-1]
    xmaxtick = ax.get_xticks()[-1]
    
    # get center and scale
    center = round(center_func(series), ndigits)
    scale = round(scale_func(series), ndigits)
    
    # plot box with center & scale
    bbox = {'fc': '1', 'pad': 3}
    ax.annotate(
    """{}: {}
{}: {}""".format(center_func.__name__,
                     round(center, ndigits),
                     scale_func.__name__,
                     round(scale, ndigits)),
        xy=(.8* xmaxtick, .8 * ymaxtick),
        bbox=bbox)

    # annotate center
    ax.annotate("{}".format(center), 
            xy=(center, 
                .8 * ymaxtick),
            xytext=(1.19 * center, 
                    .8 * ymaxtick),
            arrowprops={'facecolor': 'black', 
                        'arrowstyle': "-|>"})
    
    # plot first negative band if applicable
    if ax.get_xticks()[0] < center - scale:
        bandn1 = round(center - scale, ndigits)
        ax.axvspan(bandn1,
                   center, 
                   alpha=0.5, 
                   color='g')
        ax.axvline(center, c='b')
    else:
        ax.axvspan(ax.get_xticks()[0],
                   center, 
                   alpha=0.2, 
                   color='blue')

    # plot first band
    band1 = round(center + scale, ndigits)
    ax.axvspan(center,
               band1, 
               alpha=0.5, 
               color='g')    

    # annotate band 1
    ax.annotate("{}".format(band1), 
            xy=(band1, 
                .7 * ymaxtick),
            xytext=(band1 + .19 * center, 
                    .7 * ymaxtick),
            arrowprops={'facecolor': 'black', 
                        'arrowstyle': "-|>"})

    # plot negative second band if applicable
    if ax.get_xticks()[0] < center -  2 * scale:
        bandn2 = round(center - 2 * scale, ndigits)
        ax.axvspan(bandn2,
                   bandn1, 
                   alpha=0.5, 
                   color='y')

    # plot second band
    band2 = round(center + 2 * scale, ndigits)
    ax.axvspan(band1,
               band2, 
               alpha=0.5, 
               color='y')

    # annotate second band
    ax.annotate("{}".format(band2), 
            xy=(band2, 
                .6 * ymaxtick),
            xytext=(band2 + .19 * center, 
                    .6 * ymaxtick),
            arrowprops={'facecolor': 'black', 
                        'arrowstyle': "-|>"})
    
    # plot negative third band if applicable
    if ax.get_xticks()[0] < center -  3 * scale:
        bandn3 = round(center - 3 * scale, ndigits)
        ax.axvspan(bandn3,
                   bandn2, 
                   alpha=0.5, 
                   color='purple')

    # plot third band
    band3 = round(center + 3 * scale, ndigits)
    ax.axvspan(band2,
               band3, 
               alpha=0.5, 
               color='purple')
    
    # annotate third band
    ax.annotate("{}".format(band3), 
            xy=(band3, 
                .5 * ymaxtick),
            xytext=(band3 + .19 * center, 
                    .5 * ymaxtick),
            arrowprops={'facecolor': 'black', 
                        'arrowstyle': "-|>"})

    ax.title.set_text('{}: histogram, {} & {}'.format(
                            series.name, 
                            center_func.__name__, 
                            scale_func.__name__))
    if return_bins:
        return ax, bins
    else:
        return ax


def center_scale_with_peak_trough(series, plot_troughs=True, plot_peaks=False,
                                  bins=100, center_scale_plot_kwargs=None, 
                                  trough_plt_kwargs=None, peak_plt_kwargs=None):
    """
    Produces `center_scale_plot` but with the option to add peak and / or trough
    lines to the resulting histogram. See `features.find_frequency_peaks_and_troughs` 
    for more information on peak and trough calculation.
    """
    def plot_optima(optima, **kwargs):
        for opt in np.mean(optima, axis=1):
            
            if "linestyle" not in kwargs:
                kwargs["linestyle"] = "--"
            
            if "linewidth" not in kwargs:
                kwargs["linewidth"] = 1

            ax.axvline(opt, 0, y_max, **kwargs)
    
    if center_scale_plot_kwargs is None:
        center_scale_plot_kwargs = {"center_func": np.mean,
                                    "scale_func": np.std}
    
    if trough_plt_kwargs is None:
        trough_plt_kwargs = {}
        
    if peak_plt_kwargs is None:
        peak_plt_kwargs = {}
    
    ax = center_scale_plot(series, bins=bins, **center_scale_plot_kwargs)
    peaks, troughs = features.find_frequency_peaks_and_troughs(series, bins=bins)
    
    y_max = ax.get_ybound()[1]
    
    if plot_troughs:
        plot_optima(troughs, **trough_plt_kwargs)
    
    if plot_peaks:
        plot_optima(peaks, **peak_plt_kwargs)
    
    return ax


import numpy as np
import pandas as pd

from sklearn.metrics import precision_score, recall_score

import sklearn.model_selection as skl
import matplotlib.pyplot as plt

from indoorplants.validation import crossvalidate, \
                                    calibration, \
                                    boundaries


def validation_curve(X, y, score, model_type, param_name,
                     param_range, other_params={},
                     semilog=False, figsize=(11, 8),
                     **cv_engine_kwargs):
    """
    Cross validates `model_type` across passed parameters and
    plots results. Please see _validate_param_range for more
    details around the cross validation arguments.

    Pass True for `semilog` if `param_range` values would be better
    visualized with log scaling. Pass tuple to `figsize` if you 
    wish to override default of (11, 8).

    Please see `validation.cv_engine` for details on other args.
    """
    results = crossvalidate.validate_param_range(
                                X, y, model_type, param_name, param_range,
                                [score], other_params, **cv_engine_kwargs)

    # special handling for iterable hyper-parameter values
    if results.index.levels[0].dtype == "O":

        # map `param_range` values to `str`
        param_range = list(map(str, param_range))

        # get `Categorical` column, which will preserve original ordering
        cat_type = pd.api.types.CategoricalDtype(param_range, ordered=True)
        cat_df = pd.DataFrame(param_range, columns=[param_name]
                  ).astype(cat_type)

        # temporarily drop score function names from `results`
        score_name = results.columns.levels[0][0]
        results.columns = results.columns.droplevel(0)

        # merge in `Categorical` columns
        results = results.reset_index(
                        ).rename(columns={param_name: "param_value_str"})

        results = results.merge(cat_df, left_on="param_value_str", right_on=param_name
                        ).drop("param_value_str", axis=1)

        # sort `results` correctly and set index
        results = results.sort_values(param_name
                        ).set_index([param_name, "level_1"])

        # set `score_name` as top level column
        results = pd.concat({score_name: results}, axis=1)

    # get stats
    means = results.groupby(level=0).mean().reset_index()
    stds = results.groupby(level=0).std().reset_index()

    # create and setup figure and axes
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    xtick = ax.set_xticks(means.index)

    # handle logarithmic axes
    if semilog is True:
        plt_func = plt.semilogx
    else:
        plt_func = plt.plot

    # plot average train scores
    plt_func(means.index, means[(score.__name__, "train")].values, 
             label="train", color="darkorange", lw=2)

    # plot average validation scores
    plt_func(means.index, means[(score.__name__, "test")].values, 
             label="validation", color="navy", lw=2)

    # plot standard deviation bands around average scores
    bands = lambda _: (means[(score.__name__, _)] - stds[(score.__name__, _)],
                       means[(score.__name__, _)] + stds[(score.__name__, _)])

    plt.fill_between(means.index, *bands("train"), alpha=0.1, color="darkorange", lw=2)
    plt.fill_between(means.index, *bands("test"), alpha=0.1, color="navy", lw=2)

    # final plot formatting
    xlab = ax.set_xlabel(param_name)
    xlab = ax.set_xticklabels(means[param_name].values)
    ylab = ax.set_ylabel(score.__name__)
    title = plt.title(f"validation curve: {model_type.__name__}, across {param_name}")
    plt.legend(loc="best")


def learning_curve(X, y, model_type, score, model_params=None,
                   splits=5, scale_obj=None, random_state=0,
                   train_sizes=[0.1, 0.33, 0.55, 0.78, 1.],
                   figsize=(11, 8)):
    """
    Cross validates 'model_type' over different `train_sizes`, which allows 
    for both insight into model fit and insight into whether model might 
    benefit from access to more data. Plots results.

    Please see `validation.cv_engine` for details on other args.
    """
    if model_params is None:
        model_params = {}
    model_obj = model_type(**model_params)

    results = crossvalidate.validate_train_sizes(
                                X, y, model_obj, score_funcs=[score], 
                                splits=splits, scale_obj=scale_obj,
                                train_sizes=train_sizes,
                                random_state=random_state)

    means = results.groupby(level=0).mean().reset_index()
    stds = results.groupby(level=0).std().reset_index()

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    xtick = ax.set_xticks(means.index)

    plt.plot(means.index, means[(score.__name__, "train")
                                ].values, 
             label="train", color="crimson", lw=2)
    plt.plot(means.index, means[(score.__name__, "test")
                                ].values, 
             label="validation", color="teal", lw=2)

    bands = lambda _: (means[(score.__name__, _)]
                        - stds[(score.__name__, _)],
                       means[(score.__name__, _)]
                        + stds[(score.__name__, _)])

    plt.fill_between(means.index, *bands("train"), 
                     alpha=0.1, color="crimson", lw=2)
    plt.fill_between(means.index, *bands("test"), 
                     alpha=0.1, color="teal", lw=2)

    xlab = ax.set_xlabel("training size")
    xlab = means["index"].map(lambda _: int(round(float(_) * len(y), -2)))

    xlab = ax.set_xticklabels(xlab)
    ylab = ax.set_ylabel(score.__name__)

    title = plt.title(f"Learning curve: {model_type.__name__}")
    plt.legend(loc="best")


def calibration_curve(X, y, model_type, model_params={}, calib_types=None, 
                      figsize=(11, 8), display_counts=True, **cv_engine_kwargs):
    """
    Plots calibration curves for original model & passed calibrators.

    Please see `validation.cv_engine` for details on other args.
    """
    def plot_probs_and_counts(results, c, label, plot_counts=display_counts):
        # greb emprirical probability from `results`
        probs = results["empirical_probability"]

        # plot empirical probability
        plt.plot(probs.index, probs["mean"], label=label, color=c, lw=2)

        # function to get +/- std around empirical probability means
        bands = lambda bin: (bin["mean"] - bin["std"], bin["mean"] + bin["std"])

        # plot counts with std lines if `display_counts` is True
        if plot_counts is True:
            counts = results["proportion_of_test_data"]
            plt.bar(counts.index, counts["mean"], yerr=counts["std"],
                    color="steelblue", width=.025, alpha=.2, label="bin_count_normalized")

        # fill std around original model"s mean calibration
        plt.fill_between(probs.index, *bands(probs), alpha=0.1, color=c, lw=2)

        # display horizontal grid lines
        plt.grid(which="major", axis="y", color="grey", linestyle="--")

    model_obj = model_type(**model_params)
    results = calibration.cv_calibrate(X=X, y=y,
                                       model_obj=model_obj,
                                       calib_types=calib_types,
                                       retain_counts=display_counts,
                                       **cv_engine_kwargs)

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    xtick = ax.set_xticks(results.index)
    colors = (c for c in ["C6", "C4", "C0"])

    plot = plot_probs_and_counts(results.loc[:, "original_model"], 
                                 colors.__next__(), 
                                 "original_model")

    if calib_types is not None:
        for cal_mod in calib_types:
            plot = plot_probs_and_counts(results.loc[:, cal_mod.__name__ + "_cal"], 
                                         colors.__next__(), 
                                         cal_mod.__name__,
                                         plot_counts=False)

    ax.set_xlim(0, 1)
    xlab = ax.set_xlabel("predicted probability (bin)")

    ax.set_ylim(0, 1)
    ylab = ax.set_ylabel("empirical probability")

    title = plt.title(f"calibration curve: {model_type.__name__}")
    plt.legend(loc="best")


def precision_recall_curve(X, y, model_type, scale_obj=None, 
                           splits=5, model_params={}, figsize=(11, 8),
                           thresholds=[.1*x for x in range(1, 10)]):
    """
    Plot precision-recall curve over decision boundaries:
    [0, 1] for binary classification. 

    Pass `model_type`, `model_params` and `figsze` in same fashion 
    as for learning_curve. `splits` and `scale_obj` are same as
    for all cv_score functions.
    """
    results = boundaries.cv_score(X, y, model_type(**model_params), 
                                 [recall_score, precision_score],
                                 thresholds,
                                 splits, False, scale_obj)
    to_plot = results.unstack()["mean"]

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)

    plt.plot(to_plot["recall_score"], to_plot["precision_score"], lw=2)
    plt.fill_between(to_plot["recall_score"], to_plot["precision_score"],
                     alpha=.2)

    for row in to_plot.itertuples():
        ax.annotate("{}".format(round(row[0], 3)),
                    xy=(row[1], row[2]),
                    xytext=(row[1] - .01, 
                            row[2] + .02))
    
    plt.xlabel("pecall")
    plt.ylabel("precision")
    title = plt.title("precision & recall by decision boundary: {}".format(
                                        model_type.__name__))


import itertools
import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler


def euclidean(feature_one, feature_two):
    """
    Returns Euclidean distance between the two passed features. Both
    are assumed to be of type `pd.Series` or uni-column `pd.DataFrame`.
    """
    return np.sqrt(
                (feature_one - feature_two
                   ).pow(2
                   ).sum())


def euclidean_squared(feature_one, feature_two):
    """
    Returns squared Euclidean distance, which accentuates dispartity moreso
    than its square root does, between the two passed features. Both are 
    assumed to be of type `pd.Series` or uni-column `pd.DataFrame`.
    """
    return (feature_one - feature_two
               ).pow(2
               ).sum()


def get_feature_iter_func(flat):
    """
    Helper function to determine what type of iteration to perform across two
    features. Takes `bool` as input.
    """
    if flat:
        return lambda feature_names: itertools.combinations(feature_names, 2)
    else:
        return lambda feature_names: itertools.product(feature_names, feature_names)


def normalize_min_max(df):
    """
    Uses min-max normalization to scale passed `pd.DataFrame` of values. Returns
    `pd.DataFrame` of transformed values with original row and column labels.
    """
    cols = df.columns
    inds = df.index

    arr = MinMaxScaler().fit_transform(df)
    return pd.DataFrame(arr, columns=cols, index=inds)


def feature_distances(distance_func):
    """
    Decorator for functions that calculate feature distances.

    `distance_func` is assumed to take as input, two features, likely in
    following the terms laid out in `euclidean`.

    This function will iterate through either the Cartesian Product (with
    `flat=True`) or all combinations of 2 (`flat=False`) of the specific
    features of the passed `DataFrame`. It will then return a nicely
    formatted `DataFrame` of results.

    Description of `inner` arguments, in order:
    - `pd.DataFrame` of data
    - `feature_names_list`, a list of  columns in `df` to compare
    - `flat`, boolean denoting result display style
    - `normalize_data`, which determines whether to min-max-scale data
    """
    def inner(df, feature_names_list, flat=True, normalize_data=True):
        if normalize_data:
            df = normalize_min_max(df[feature_names_list].astype(float))

        iter_func = get_feature_iter_func(flat)

        scores = [(f1, f2, distance_func(df[f1], df[f2]))
                  for f1, f2 in iter_func(feature_names_list)]

        scores_df = pd.DataFrame(scores, columns=["feature_one",
                                                  "feature_two",
                                                  f"{distance_func.__name__}"])

        if flat:
            scores_df = scores_df.set_index(["feature_one",
                                             "feature_two"]
                                ).sort_values(f"{distance_func.__name__}", 
                                              ascending=False)

        else:
            scores_df = scores_df.pivot(index='feature_one', 
                                        columns='feature_two', 
                                        values=f"{distance_func.__name__}")

        return scores_df

    return inner


@feature_distances
def euclidean_distances(feature_one, feature_two, **kwargs):
    """
    Wrapper allowing for running `euclidean` across many feature pairs.

    See `feature_distances` for more detail on how to pass arguments,
    and `euclidean` for details on metric calculation.
    """
    return euclidean(feature_one, feature_two)


@feature_distances
def euclidean_squared_distances(feature_one, feature_two, **kwargs):
    """
    Wrapper allowing for running `euclidean_squared` across many feature pairs.

    See `feature_distances` for more detail on how to pass arguments,
    and `euclidean_squared` for details on metric calculation.
    """
    return euclidean_squared(feature_one, feature_two)


def get_sign_diffs(array_like):
    """
    Given `array_like` (`np.array` or `pd.Series`), return an 
    array noting the change in direction from `array_like[row_index]` 
    to `array_like[row_index + 1]` for each row index in `array_like`.
    """
    return np.diff(np.sign(np.diff(array_like)))


def get_optima_indices_simple(sign_diffs_arr, sign_change_val):
    """
    Get inidices for peaks or troughs in passed `sign_diffs_arr`. Note that 
    `sign_change_val` should be -2 for peaks or 2 for troughs.
    """
    return np.flatnonzero(sign_diffs_arr == sign_change_val) + 1


def get_peak_and_trough_indices_simple(array_like):
    """
    Get indices of peaks and troughs in `array_like` using simple methodology 
    of looking for changes in direction. This method will likely benefit from 
    smoothing if data is "noisey."
    """
    sign_diffs_arr = get_sign_diffs(array_like)

    return (
        get_optima_indices_simple(sign_diffs_arr, -2),
        get_optima_indices_simple(sign_diffs_arr, 2)
        )


def get_bin_edge_inds_from_hist_inds(indices, bin_edges):
    """
    Given an array of indices (i.e. integer values - this was built with  
    the results of `get_optima_indices_simple` in mind) and an array of 
    histogram bin edges (assumed to be the second result of `np.histogram`) 
    return a `[num_bins, 2]` array, such that `bin_edges` is re-represented 
    to provide exactly two boundaries, with each pair corresponding to exactly 
    one value in `indices`. 
    """
    return bin_edges[np.vstack((indices, indices + 1)).T]


def find_frequency_peaks_and_troughs(series=None, hist=None, 
                                     bin_edges=None, bins=100):
    """
    Leverage functionality in `get_peak_and_trough_indices_simple` on 
    histogram data. If not passed (in `hist` and `bin_edges`), histogram 
    data is generated with `num_bins` bins.
    """
    if hist is None or bin_edges is None:
        hist, bin_edges = np.histogram(series, bins=bins)

    hist_peak_inds, hist_trough_inds = get_peak_and_trough_indices_simple(hist)

    return (
        get_bin_edge_inds_from_hist_inds(hist_peak_inds, bin_edges),
        get_bin_edge_inds_from_hist_inds(hist_trough_inds, bin_edges)
        )


def get_redundant_feature_pairs(feature_distance_series, threshold):
    """
    Expects results from a `feature_distances` func. Returns redundant 
    feature pairs, as determined by passed `threshold` for inter-feature 
    measurement.
    """
    return feature_distance_series[
                    feature_distance_series < threshold
                    ].reset_index(
                    ).iloc[:, :2
                    ].drop_duplicates()


import pandas as pd


def get_clean_df_index(df):
    if isinstance(df.index, pd.core.index.MultiIndex):
        return df.index.remove_unused_levels()
    else:
        return df.index


def get_feature_size_by_class(df, class_col, features, normalize="class"):
    """
    Given pandas.DataFrame, class column name, and feature column
    name, return:

        - pd.crosstab(df.class_col, df.feature).stack()

    Works differently from `pandas.crosstab` if multiple features passed.
    Note that this function makes the most sense for categorical 
    features.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    class_col : str
        Column name for the class / target.

    features : str or list
        Column name(s) for the feature.

    normalize : str, "class" or "all" or None (default="class")
        Whether to normalize counts. "class" means normalize counts within a 
        class value by the total counts in that group. "all" means normalize 
        by the length of `df`. `None` means do not normalize.

    Return
    ------

    pandas.DataFrame of size figures.
    """
    if isinstance(features, str):
        features = [features]
    elif not isinstance(features, list):
        raise TypeError("Must pass str or list for `features`.")

    to_group_by = [class_col] + features

    to_return = df[to_group_by
                  ].groupby(to_group_by
                  ).size(
                  ).rename("cnt"
                  ).to_frame()

    if normalize == "all":
        to_return = (to_return / len(df)
                    ).rename(columns={"cnt": "ratio"})

    elif normalize == "class":
        denominators = to_return.groupby(level=0
                               ).sum(
                               ).rename(columns={"cnt": "cnt_cls_val"}
                               ).reset_index()

        to_return = to_return.reset_index(
                            ).merge(denominators, on=class_col
                            ).set_index(to_group_by)

        to_return = to_return.cnt.divide(to_return.cnt_cls_val
                                ).rename("ratio"
                                ).to_frame()

    elif normalize is None:
        pass

    else:
        raise ValueError("Improper value passed for `normalize`.")

    return to_return


def get_feature_sizes_dict(df, class_col, normalize="class"):
    """
    Helper function to get size of each feature-value & class-value set,
    with `class_col` excluded from feature set to be examined.
    """
    return {
                feature: get_feature_size_by_class(df,
                                                   class_col,
                                                   feature,
                                                   normalize=normalize)

                for feature in filter(lambda c: c != class_col, df.columns)
    }


def get_class_cnts_by_feature_null(df, class_col, feature, normalize=True):
    """
    Break out class fequencies (in `df[class_col]`) by whether or not 
    `df[feature]` is null.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    class_col : str
        Column name for the class / target.

    feature : str
        Column name for the feature.

    normalize : bool (default=True)
        Whether or not to normalize class counts by number of rows in 
        the respective feature is: [null / non-null] query. I.e. the
        value for `normalize` is passed straight to the `normalize`
        kwarg in `pandas.Series.value_counts`, which is called on data that
        is filtered for either `df[feature].isnull()` of `df[feature].notnull()`.

    Return
    ------

    pandas.DataFrame of class counts, broken out by whether or not 
    `df[feature]` is null.
    """
    null = df.loc[df[feature].isnull(), class_col
                 ].value_counts(normalize=normalize
                 ).rename("null"
                 ).to_frame()

    not_null = df.loc[df[feature].notnull(), class_col
                     ].value_counts(normalize=normalize
                     ).rename("not_null")

    return pd.concat({feature: null.join(not_null)}, axis=1)


def get_class_cnts_by_many_features_nulls(df, class_col, features_list,
                                          normalize=True):
    """
    Wrapper for `get_class_cnts_by_feature_null`, with only difference being 
    that, in this function, users passes a `list` of features, as opposed to 
    a single feature.
    """ 
    return pd.concat(
        [
            get_class_cnts_by_feature_null(df, class_col, f, normalize=normalize)
            for f in features_list
        ],
        axis=1,
        sort=True
    ).T


def get_null_stats(df):
    """
    Function that will return count (absolute and relative)  of null 
    values in each column in `df`.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    Return
    ------

    pandas.DataFrame with `df.columns` as row index and `["cnt", "ratio"]` 
    (for absolute and relative counts, respectively) in column index.
    """
    nulls = df.isnull().sum()
    nulls = nulls.rename("cnt").to_frame()
    nulls["ratio"] = nulls / len(df)
    return nulls.sort_values("ratio", ascending=False)


def get_cols_over_x_pcnt_null(df, x=.99, exclude=None, names_only=True):
    """
    Function that will return all columns in `df` with over a certan 
    proportion of missing values.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    x : float, with value between 0 and 1 (default=.99)
        Proportion of rows allowed to be missing in a given column.

    exclude : str or list (default=None)
        Column name(s) to exclude from results.

    names_only : bool (default=True)
        Whether to return just column names (True) or whether to return 
        `get_null_stats` results for these columns (False).

    Return
    ------

    `pandas.Index` of columns with over `x` proportion missing values, or 
    `pandas.DataFrame` of `null` ratios.
    """
    nulls = get_null_stats(df)

    if exclude:
        if isinstance(exclude, str):
            exclude = [exclude]

        nulls = nulls.loc[~nulls.index.isin(exclude)]

    if names_only:
        return get_clean_df_index(nulls[nulls.ratio > x])
    else:
        return nulls[nulls.ratio > x]


def remove_cols_over_x_pcnt_null(df, x=.99, exclude=None):
    """
    Function that will remove from `df` all columns with over an `x` proportion 
    of missing values. User can use `exclude` to prevent certain columns from 
    being removed.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    x : float, with value between 0 and 1 (default=.99)
        Proportion of rows allowed to be missing in a given column.

    exclude : str or list (default=None)
        Column name(s) to exclude from removal (i.e. for use when you know a
        given column has missing values but you'd like to keep it anyway).

    Return
    ------

    pandas.DataFrame with columns with columns with over an `x` proportion 
    of missing values removed.
    """
    to_remove = get_cols_over_x_pcnt_null(df, x=x, exclude=exclude)
    return df.drop(to_remove, axis=1)


def make_is_null_cols(df, x=.5, exclude=None, remove_originals=False):
    """
    Function that will make in `df` boolean re-representaions of all 
    columns with over an `x` proportion of missing values. I.e., allows 
    using simply whether the value of that column in that row is `null` 
    instead of taking the actual column value.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate.

    x : float, with value between 0 and 1 (default=.99)
        Proportion of rows allowed to be missing in a given column.

    exclude : str or list (default=None)
        Column name(s) to exclude from consideration (i.e. for use when you 
        know a given column has missing values but you'd like to keep it 
        anyway).

    remove_originals : bool (default=False)
        Determines whether original columns are removed or left in `df`.

    Return
    ------

    `df`, including new, `is_null_...` columns.
    """
    null_cols = get_cols_over_x_pcnt_null(df, x=x, exclude=exclude)

    for col in null_cols:
        df["is_null_" + col] = df[col].isnull()

    if remove_originals:
        df = df.drop(null_cols, axis=1)

    return df


def get_cols_ratio_equal_val(df, val, ratio=1):
    """
    Useful for finding cols where a certain proportion of rows are equal
    to a particular value.
    """
    check = (df == val).sum()
    return check[check == (ratio * len(df))]


def is_feature_not_present_across_class(feature_size_by_class_df):
    """
    Helper function to see whether all subsections in `feature_size_by_class_df`, 
    based on slicing with different values in first-level rwo index, have same 
    number of values in second-level row index.

    Note that `feature_size_by_class_df` then must obviously have at least two 
    row-index levels.

    The intended use case is for handling results from `get_feature_size_by_class`,
    where first-level row index corresponds to class values, and second-level
    row index corresponds to feature values.

    The idea is to confirm that all feature values are present across all class 
    values. If this is not true, feature is either highly predictive or leaking
    data.

    Feature in `feature_size_by_class_df` should be discrete, i.e. of a finite 
    set of values.

    Parameters
    ----------

    feature_size_by_class_df : pandas.DataFrame
        DataFrame on which this function will operate, which must have at least
        two row-index levels.

    Return
    ------

    `bool` indicating whether all class values see all feature values.
    """
    index_vals = feature_size_by_class_df.reset_index().iloc[:, :2]
    counts = index_vals.groupby(index_vals.columns[0]).size()
    return (
        counts.nunique() != 1  
            or counts.index.nunique() != 
               feature_size_by_class_df.index.levels[0].nunique()
            )


def is_feature_overweighted_towards_class(feature_size_by_class_df,
                                          threshold=.99,
                                          feature_level=True):
    """
    The intended use case is for seeing whether a categorical feature very 
    strongly points towards a particular class value, as, if this is true, 
    the feature is either highly predictive or leaking data.

    Parameters
    ----------

    feature_size_by_class_df : pandas.DataFrame
        DataFrame on which this function will operate. This is presumed to 
        be a result from the function `get_feature_size_by_class`. Additionally, 
        the feature represented in `feature_size_by_class_df` should be discrete, 
        i.e. comes from a finite set of values.

    threshold : float, with value between 0 and 1 (default=.99)
        Proportion of rows allowed that are allowed to be taken up by single
        `class_col` & `feature` value grouping.

    feature_level : bool (default=True)
        Whether to perform at the feature level, or feature-value level.

    Return
    ------

    `bool` indicating whether all class values see all feature values.
    """
    ratios = feature_size_by_class_df.unstack(0).ratio

    if feature_level:
        booleans = ratios.sum() >= threshold
    else:
        booleans = (ratios >= threshold).any()

    return booleans.any()


def get_data_leak_cols_via_cls_counts(df, class_col, threshold=.99,
                                      dtypes=None, drop_for_analysis=None,
                                      join_for_analysis=None,
                                      return_style="dict"):
    """
    Function to assess whether features (categorical) in `df` are possibly 
    leaking data. This is achieved through through determining whether, for 
    a given feature:

        - every class value sees the same number of feature values
        - whether a disproportionate number of rows correspond to a certain 
          class value (where disproportionality is set using `threshold`)

    This latter insight is provided both at the feature and feature-value 
    level. See the `is_feature_overweighted_towards_class` function, or 
    below desctiption of `return_style` kwarg for more information.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate - features should be 
        categorical.

    class_col: str
        Name of class column.

    threshold : float, with value between 0 and 1 (default=.99)
        Proportion of rows allowed that are allowed to be taken up by single
        `class_col` & `feature` or `class_col` & `feature`-value grouping.

    dtypes : type, or list[type] (default=None)
        Specific column type(s) to limit analysis to. Defaults to all types 
        except `float` if no value passed.

    drop_for_analysis : str or list (default=None)
        Column name(s) to exclude from consideration.

    join_for_analysis : pd.Series or pd.DataFrame (default=None)
        Additional data to be joined in to `df; e.g. if separate `X` and `y`.
        `class_col` will be looked up post-join, if there is a join.

    return_style : str (default="dict")
        If "dict" is passed, the following groups will be returned separately, 
        each under their own key in a `dict`:

            - not all class values see all feature values ("missing_vals")
            - there is a disproportionate concentration of a certain class 
              value at a certain feature ("overweight_feature")
            - there is a disproportionate concentration of a certain class 
              value at a certain feature value ("overweight_feature_value")

        Otherwise, a single `list` will be returned that contains all 
        column names.

    Return
    ------

    Columns with potential data leak issues, formatted as per `return_style`.
    """
    if dtypes is None:
        kwargs = {"exclude": float}
    else:
        kwargs = {"include": dtypes}

    if drop_for_analysis:
        df = df.drop(drop_for_analysis, axis=1)

    df = df.select_dtypes(**kwargs)

    if join_for_analysis is not None:
        df = df.join(join_for_analysis)

    feature_sizes = get_feature_sizes_dict(df, class_col, normalize="class")

    filter_feature_data = lambda func, *args, **kwargs: [
                            k for k, v in feature_sizes.items()
                            if func(v, *args, **kwargs)
                        ]

    missing_vals = filter_feature_data(
                                is_feature_not_present_across_class
                                )

    overweight_feature = filter_feature_data(
                                is_feature_overweighted_towards_class, 
                                threshold,
                                feature_level=True
                                )

    overweight_feature_val = filter_feature_data(
                                is_feature_overweighted_towards_class,
                                threshold,
                                feature_level=False
                                )

    if return_style == "dict":
        return {
            "missing_vals": missing_vals,
            "overweight_feature": overweight_feature,
            "overweight_feature_val": overweight_feature_val
        }

    else:
        return list(
                    set(missing_vals)
                  | set(overweight_feature)
                  | set(overweight_feature_val)
                )


def get_missing_around_class_stats(null_counts, names_only=True):
    """
    Given results from `get_class_cnts_by_many_features_nulls`, get
    features whose values are entirely `null` at a given class 
    value.

    Parameters
    ----------

    null_counts : pandas.DataFrame
        Results from `get_class_cnts_by_many_features_nulls`.

    names_only : bool (default=True)
        Whether to return just names of features who are `null` for 
        a given class value (True) or whether to return full `null_counts` 
        table for these features (False).

    Return
    ------

    Either `pandas.DataFrame` or `list`, representing `null_counts` filtered 
    to features that are entirely `null` at a given class value.
    """
    missing_vals = null_counts[null_counts.isnull().any(axis=1)
                              ].dropna(how="all")

    if names_only:
        missing_vals = list(
                        get_clean_df_index(missing_vals
                            ).levels[0]
                        )

    return missing_vals


def get_null_count_spreads(null_counts, exclude=None):
    """
    Given results from `get_class_cnts_by_many_features_nulls`, get absolute 
    values of differences, for each class value, between the proportion of rows 
    taking that class value when a given feature is `null` and when a given 
    feature is `not-null`.

    Parameters
    ----------

    null_counts : pandas.DataFrame
        Results from `get_class_cnts_by_many_features_nulls`.

    exclude : list-like (default=None)
        Features to remove from returned results.

    Return
    ------

    `pandas.DataFrame` with features as indices, class values as columns, and 
    `not-null` absolute differences (as described above) as values.
    """
    if exclude is None:
        exclude = []

    null_counts.index = null_counts.index.rename(["feature", "is_null"])
    null_counts = null_counts.reset_index()

    null_counts = null_counts.loc[
                            (~null_counts.feature.isin(exclude))
                          & (null_counts.notnull().all(axis=1))
                         ]

    to_return = null_counts.groupby("feature"
                          ).diff(
                          ).set_index(null_counts.feature
                          ).dropna(
                          ).abs(
                          ).iloc[:, 0]

    to_return.columns = ["spread"]
    return to_return


def get_over_threshold_columns(spreads, threshold, names_only=True):
    """
    Given results from `get_null_count_spreads`, get features whose 
    spreads exceed `threshold`.

    Parameters
    ----------

    spreads : pandas.DataFrame
        Results from `get_null_count_spreads`.

    threshold : float, with value between 0 and 1 (default=.99)
        Measure of maximum acceptibility for spread between the proportion 
        of rows with a given class value when a given feature is `null`, vs. 
        when it's `not-null`. Spread is the difference between these two 
        proportions.

    names_only : bool (default=True)
        Whether to return just names of features that exceed `threshold`, or 
        the `get_null_count_spreads` results for these features.

    Return
    ------

    Either `pandas.DataFrame` or `list`.
    """
    over_threshold = spreads[spreads > threshold].dropna()

    if names_only:
        spreads = list(
                    spreads[spreads > threshold
                          ].dropna(
                          ).index
                )

    return spreads


def get_data_leak_cols_via_nulls(df, class_col, threshold=.5, dtypes=float,
                                 drop_for_analysis=None, join_for_analysis=None,
                                 return_style="dict"):
    """
    Function to assess whether features in `df` are possibly leaking data. 
    This is achieved in a simple fashion, through considering the presence of 
    `null` values and how that varies across different class values. More 
    specically, this function provides:

        - whether a feature is entirely `null` over a certain class value
        - whether there's a unreasonably high spread between the `null` and 
          `not-null` counts at a given class value, where the maximim spread 
          considered reasonable is set using `threshold`

    Note that these two checks are mutually exclusive, and also that 
    features whose values are not missing at all will be excluded from the
    results.

    Parameters
    ----------

    df : pandas.DataFrame
        DataFrame on which this function will operate. These features can 
        be categorical or continuous.

    class_col : str
        Name of class column.

    threshold : float, with value between 0 and 1 (default=.99)
        Measure of maximum acceptibility for spread between the proportion 
        of rows with a given class value when a given feature is `null`, vs. 
        when it's `not-null`. Spread is the difference between these two 
        proportions.

    dtypes : type, or list[type] (default=float)
        Specific column type(s) to limit analysis to.

    drop_for_analysis : str or list (default=None)
        Column name(s) to exclude from consideration.

    join_for_analysis : pd.Series or pd.DataFrame (default=None)
        Additional data to be joined in to `df; e.g. if separate `X` and `y`.
        `class_col` will be looked up post-join, if there is a join.

    return_style : str (default="dict")
        If "dict" is passed, the analysis results will be returned separately, 
        each under their own key in a `dict`:

            - features for which one or more class values see only `null`
              ("missing_vals")
            - features for which spread, as defined above, exceeds `threshold` 
              ("over_threshold")

        Otherwise, a single `list` will be returned that contains all 
        column names.

    Return
    ------

    Columns with potential data leak issues, formatted as per `return_style`.
    """
    if drop_for_analysis:
        df = df.drop(drop_for_analysis, axis=1)

    df = df.select_dtypes(include=dtypes)

    if join_for_analysis is not None:
        df = df.join(join_for_analysis)

    null_counts = get_class_cnts_by_many_features_nulls(
                                df,
                                class_col,
                                filter(lambda c: c != class_col, df.columns)
                                )

    missing_vals = get_missing_around_class_stats(null_counts)

    over_threshold = get_over_threshold_columns(
                        spreads=get_null_count_spreads(
                                        null_counts=null_counts, 
                                        exclude=missing_vals
                                        ),
                        threshold=threshold
                        )

    if return_style == "dict":
        return {
                "missing_vals": missing_vals,
                "over_threshold": over_threshold
            }

    else:
        return missing_vals + over_threshold


import numpy as np
import pandas as pd
from itertools import chain

from indoorplants.validation import crossvalidate


def _cv_proba(X, y, model_obj, splits=5, **cv_engine_kwargs):
    """
    Return numpy array of "splits"-fold CV results, 
    where results are:
    -column 0: actual class
    -columns 1 & 2: proba. of each of neg. and pos. classes.
    """
    model_obj.predict = model_obj.predict_proba

    score = lambda true, predicted: np.hstack(
                [true.values.reshape(-1, 1), predicted[:, 1].reshape(-1, 1)]
                )

    return crossvalidate.cv_engine(
                            X=X, y=y,
                            model_obj=model_obj,
                            score_funcs=[score],
                            splits=splits,
                            train_scores=False,
                            **cv_engine_kwargs
                            )


def _get_rank_stats(results):
    """
    Helper function that produces median and m.a.d
    probabilities for each of 2 classes given _cv_proba 
    results with reworked columns.
    """
    grouped = results[["class", "proba"]
                     ].groupby(
                        ["class"]
                     ).median(
                     ).rename(columns={"proba":"median"})

    return grouped.join(results[
                            ["class", "proba"]
                                ].groupby(["class"]
                  ).mad(
                  ).rename(columns={"proba":"mad"}))


def cv_rank(X, y, model_obj, splits=5, **cv_engine_kwargs):
    """
    Returns median and m.a.d. probabilities for each class
    for test results over "splits"-fold CV.
    """
    results = pd.DataFrame(
                np.vstack(
                    chain(
                        *_cv_proba(
                            X=X, y=y,
                            model_obj=model_obj,
                            splits=splits,
                            **cv_engine_kwargs
                    ))),
                columns=["class", "proba"])

    return _get_rank_stats(results)


def _calibrate_cv(model_results, calib_type, splits):
    """
    Produces calibrated probabilities for passed "model_results"
    using passed "calib_type" calibration model type 
    (uninstantiated model object).
    """
    def calibrator_fit(split_number):
        X = model_results.loc[split_number, "proba"].to_frame()
        y = model_results.loc[split_number, "class"]
        return calib_type().fit(X, y)

    def calibrator_predict_proba(calib_obj, split_number):
        X = model_results.loc[split_number, "proba"].to_frame()
        return calib_obj.predict_proba(X)

    def get_results_for_split(split_number):
        calib_obj = calibrator_fit(split_number)
        return calibrator_predict_proba(calib_obj, split_number)

    calib_res = [np.hstack(
                    [
                        model_results.loc[i, "class"].values.reshape(-1, 1),
                        get_results_for_split(i)[:, 1].reshape(-1, 1)
                    ])
                 for i in range(splits)]

    df = pd.concat({
            i: pd.DataFrame(data) for i, data in enumerate(calib_res)
            })

    df.columns = ["class", "proba"]
    return df


def _cv_calibrate(X, y, model_obj, splits=5, calib_types=None, **cv_engine_kwargs):
    """
    Returns CV results for passed model, and calibrates results.
    """
    model_res = _cv_proba(
                    X=X, y=y,
                    model_obj=model_obj,
                    splits=splits,
                    **cv_engine_kwargs
                    )

    df = pd.concat({
            i: pd.DataFrame(data) for i, data in enumerate(chain(*model_res))
        })

    df.columns = ["class", "proba"]

    if calib_types is None:
        return pd.concat({"original_model": df}, axis=1)

    else:
        dfs_cal = pd.concat(
                    [
                        pd.concat({
                            cal_type.__name__ + "_cal": _calibrate_cv(df, cal_type, splits)
                            }, axis=1) 
                        for cal_type in calib_types
                    ])

        return pd.concat({"original_model": df}, axis=1).join(dfs_cal)


def _custom_round_to_int(x, base=5):
    return int(base * round(float(x)/base))


def _custom_round_to_five_hundredth(x):
    return _custom_round_to_int(100 * x) / 100


def _prob_bin_stats(results, retain_counts=True):
    """
    Bins passed probabilities and calculates mean and std. 
    actual positive class frequencies.
    """
    df = results.copy()
    df["prob_bin"] = df.proba.apply(_custom_round_to_five_hundredth)

    # group `df` by fold, class, and probablity bin
    grouped = df[["class", "prob_bin"]
                 ].groupby(
                    [df.index.get_level_values(0), # fold
                     "class",
                     "prob_bin"]
                 ).size()

    grouped = grouped.to_frame().rename(columns={0: "count"})

    grouped = grouped.unstack(1) # pop `class` out from row- to column-level-index
    grouped.columns = grouped.columns.droplevel() # drop `count` column above `class` numbers
    grouped = grouped.drop(0, axis=1) # retain only positive class
    grouped.columns = ["count_positive"] # update columns

    # group `df` by fold and probablity bin
    grouped_other = df[["prob_bin"]
                       ].groupby(
                          [df.index.get_level_values(0), # fold
                           "prob_bin"]
                       ).size()

    grouped_other = grouped_other.to_frame().rename(columns={0: "count_bin"})

    # join the two groupings on fold and probability bin
    joined = grouped.join(grouped_other)
    joined = joined.fillna(0)

    # calculate empirical probabilities
    joined['proportion_positive_of_bin'] = joined['count_positive'] / joined['count_bin']
    joined_reset = joined.reset_index().rename(columns={"level_0": "fold"})

    # handle counts
    if retain_counts is True:
        fold_totals = joined.groupby(level=0 # get total rows per fold
                   ).sum(
                   ).loc[:, "count_bin"
                   ].rename("fold_total"
                   ).reset_index(
                   ).rename(columns={"index": "fold"})

        # merge in, and scale bin counts by, fold totals
        joined_reset = joined_reset.merge(fold_totals, on="fold")
        joined_reset["proportion_bin_of_fold"] = joined_reset["count_bin"] / joined_reset["fold_total"]

        # select final columns
        cols_to_retain = ["prob_bin", "proportion_positive_of_bin", "proportion_bin_of_fold"]
        display_columns = ["empirical_probability", "proportion_of_test_data"]

    else:
        # select final columns
         cols_to_retain = ["prob_bin", "proportion_positive_of_bin"]
         display_columns = ["empirical_probability"]

    # aggregate results and return
    means = joined_reset[cols_to_retain].groupby("prob_bin").mean()
    means.columns = display_columns
    means = pd.concat({"mean": means}, axis=1)
    means.columns = means.columns.swaplevel()

    stds = joined_reset[cols_to_retain].groupby("prob_bin").std()
    stds.columns = display_columns
    stds = pd.concat({"std": stds}, axis=1)
    stds.columns = stds.columns.swaplevel()

    to_return = means.join(stds)
    to_return.index = to_return.index.rename("predicted_probability_bin")

    return to_return


def cv_calibrate(X, y, model_obj, splits=5, calib_types=None, retain_counts=True, **cv_engine_kwargs):
    """Return mean and std. CV results comparing actual positive
    class probabilities to binned predicted probabilities, for 
    the original model and all passed calibrators."""
    res = _cv_calibrate(
                    X=X, y=y,
                    model_obj=model_obj,
                    splits=splits,
                    calib_types=calib_types,
                    **cv_engine_kwargs
                    )

    modeled = pd.concat({"original_model":
                _prob_bin_stats(res.loc[:, "original_model"], retain_counts=retain_counts)
                        }, axis=1)

    if calib_types is None:
        return modeled

    else:
        calib_results = [pd.concat(
                                {
                                    mod.__name__ + "_cal":
                                        _prob_bin_stats(res.loc[:, mod.__name__ + "_cal"],
                                                        retain_counts=retain_counts)
                                }, axis=1) 
                        for mod in calib_types]
        return modeled.join(pd.concat(calib_results))


from setuptools import setup, find_packages

setup(
	name="indoorplants",
    version="1.1",
    description="Tools for data analysis and model validation",
    author="Abe Lerman",
    url="https://github.com/lermana/indoorplants",
    packages=find_packages(),
    install_requires=["numpy",
                      "scipy",
                      "pandas",
                      "sklearn",
                      "matplotlib",
                      "requests"]
    )